# SPDX-FileCopyrightText: 2022 Henrik Sandklef
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging

from flict.flictlib.compatibility import CompatibilityFactory
from flict.flictlib.compatibility import CompatibilityStatus
from flict.flictlib.compatibility import LICENSE_COMPATIBILITY_AND
from flict.flictlib.compatibility import LICENSE_COMPATIBILITY_OR
from flict.flictlib.compatibility import CompatibilityLicenseChooser
from flict.flictlib.compatibility import CustomLicenseChooser
from flict.flictlib.alias import Alias
from flict.flictlib.license import License
from flict.flictlib.return_codes import FlictError, ReturnCodes

COMPATIBILITY_TAG = "compatibility"


class LicenseCompatibilty:

    def __init__(self, license_db=None, licenses_preferences=None, denied_licenses=None, alias_file=None):
        self.alias = Alias(alias_file)

        self.license = License(denied_licenses, self.alias)

        self.compatibility = CompatibilityFactory.get_compatibility(self.alias, license_db)

        if not licenses_preferences or licenses_preferences == []:
            self.license_chooser = CompatibilityLicenseChooser(self.compatibility.supported_licenses())
        else:
            self.license_chooser = CustomLicenseChooser(licenses_preferences)

    def inbound_outbound_compatibility(self, outbound, inbound):
        return self.compatibility.check_compat(self.alias.replace_aliases(outbound), self.alias.replace_aliases(inbound))

    def inbounds_outbound_compatibility(self, outbound, expr):
        """
        Checks how the outbound license is compatible to the inbound licenses(s).

        Parameters:
            outbound - outbound license (str)
            expr - inbound license expression (list)

        Examples:
        ----------------------------------------------------------------------
        inbounds_outbound_compatibility(X11, ['MIT', 'OR', 'LGPL-2.1-only'])
        inbounds_outbound_compatibility(X11, ['MIT OR LGPL-2.1-only']

        both return the following
            {
                "type": "operator",
                "name": "OR",
                "operands": [
                    {
                        "type": "license",
                        "name": "LGPL-2.1-only",
                        "license_aliased": "LGPL-2.1-only",
                        "check": "inbounds_outbound",
                        "outbound": {
                            "type": "license",
                            "name": "X11"
                        },
                        "outbound_aliased": "X11",
                        "allowed": true,
                        "compatibility": "No"  <--- compatiblity status for outbound X11 and inbound LGPL-2.1-only
                    },
                    {
                        "type": "license",
                        "name": "MIT",
                        "license_aliased": "MIT",
                        "check": "inbounds_outbound",
                        "outbound": {
                            "type": "license",
                            "name": "X11"
                        },
                        "outbound_aliased": "X11",
                        "allowed": true,
                        "compatibility": "Yes"  <--- compatiblity status for outbound X11 and inbound MIT
                        "problems": [] <----- list of causes such as "Check dependency, Unknown"
                    }
                ],
                "outbound": {
                    "type": "license",
                    "name": "X11"
                },
                "compatibility": "Yes",   <--- compatiblity status for outbound X11 and inbound 'MIT OR LGPL-2.1-only'.
                "allowed": true,
                "check": "inbounds_outbound"
                "problems": [] <----- list of causes such as "Check dependency, Unknown"
            }

        """
        parsed_outbound = self.license.get_license([outbound])

        logging.debug(f"inbounds_outbound_check({outbound}, {expr})")
        parsed = self.license.get_license(expr)
        logging.debug(f"inbounds_outbound_check: parsed:{parsed}")

        return self._inbounds_outbound_check(parsed_outbound, parsed)

    def __internal_expr_to_str(self, expr):
        if self.license.is_license(expr):
            return expr['name']
        elif self.license.is_operator(expr):
            raise FlictError(ReturnCodes.RET_INTERNAL_ERROR,
                             f'Internal error. Cannot transform {expr} to a license expression')

    def _inbounds_outbound_check_operator(self, outbound, expr):
        compat_summary = None
        allowed_summary = None
        problem_summary = []
        op = self.license.operator(expr)
        operands = self.license.operands(expr)
        for operand in operands:
            problems = []
            logging.debug(f"Check operand: {operand}")

            # get compatibility_tag between the operand and the outbound
            # and calculate and store the summarized compatibility
            compat = self._inbounds_outbound_check(outbound, operand)
            compat_tag = compat[COMPATIBILITY_TAG]
            if compat_tag == "Yes" or compat_tag == "No":
                compat_summary = self._update_compat(op, compat_summary, compat_tag == CompatibilityStatus.LICENSE_COMPATIBILITY_COMPATIBLE.value)
            elif compat_tag == "Unknown":
                problems.append(f'Unknown license compatibility between outbound \'{outbound["name"]}\' and {self.__internal_expr_to_str(operand)}')
                compat_summary = self._update_compat(op, compat_summary, compat_tag == CompatibilityStatus.LICENSE_COMPATIBILITY_COMPATIBLE.value)
            elif compat_tag.startswith("Check"):
                problems.append(f'Manually check license compatibility between {outbound}')
                compat_summary = self._update_compat(op, compat_summary, compat_tag == CompatibilityStatus.LICENSE_COMPATIBILITY_COMPATIBLE.value)
            elif compat_tag == "Undefined":
                raise FlictError(ReturnCodes.RET_INVALID_EXPRESSSION,
                                 'Undefined license')

            # are licenses allowed or denied for this expression
            allowed = compat['allowed']
            allowed_summary = self._update_allowed(op, allowed_summary, allowed)
            operand['allowed'] = allowed
            operand[COMPATIBILITY_TAG] = compat[COMPATIBILITY_TAG]
            operand["problems"] = problems
            problem_summary.append(problems)

        # store outbound to make for easier reading of result
        expr['outbound'] = outbound
        expr[COMPATIBILITY_TAG] = compat_summary
        expr["allowed"] = allowed_summary
        expr['check'] = 'inbounds_outbound'
        expr["problems"] = problem_summary

        return expr

    def _inbounds_outbound_check_license(self, outbound, expr):
        license_expr = self.license.license_name(expr)

        outbound_aliased = self.license.replace_aliases(self.license.license_name(outbound))
        license_aliased = self.license.replace_aliases(license_expr)

        compat = self.compatibility.check_compat(outbound_aliased, license_aliased)

        expr['license_aliased'] = license_aliased
        expr['check'] = 'inbounds_outbound'
        expr['outbound'] = outbound
        expr['outbound_aliased'] = outbound_aliased
        expr['allowed'] = self.license.license_allowed(license_expr)
        expr['problems'] = []

        expr[COMPATIBILITY_TAG] = compat[COMPATIBILITY_TAG]
        return expr

    def _inbounds_outbound_check(self, outbound, expr):
        logging.debug(f"_inbounds_outbound_check({outbound}, {expr})")
        if self.license.is_license(expr):
            return self._inbounds_outbound_check_license(outbound, expr)
        elif self.license.is_operator(expr):
            return self._inbounds_outbound_check_operator(outbound, expr)
        else:
            raise FlictError(ReturnCodes.RET_INTERNAL_ERROR,
                             f"Could not parse one of the expression: {outbound}, {expr}")

    def _update_compat(self, op, current, new):
        if current is None:
            updated = new
        elif op == LICENSE_COMPATIBILITY_AND:
            updated = current == CompatibilityStatus.LICENSE_COMPATIBILITY_COMPATIBLE.value and new
        elif op == LICENSE_COMPATIBILITY_OR:
            updated = current == CompatibilityStatus.LICENSE_COMPATIBILITY_COMPATIBLE.value or new

        if updated:
            return CompatibilityStatus.LICENSE_COMPATIBILITY_COMPATIBLE.value
        return CompatibilityStatus.LICENSE_COMPATIBILITY_INCOMPATIBLE.value

    def _update_allowed(self, op, current, new):

        if current is None:
            updated = new
        elif op == LICENSE_COMPATIBILITY_AND:
            updated = current and new
        elif op == LICENSE_COMPATIBILITY_OR:
            updated = current or new

        return updated

    def licenses(self, expr):
        return self.license.licenses(expr)

    def supported_licenses(self):
        return self.compatibility.supported_licenses()

    def replace_aliases(self, expr):
        return self.alias.replace_aliases(expr)

    def check_compatibilities(self, licenses, check_all=False):
        return self.compatibility.check_compatibilities(licenses, check_all)

    def extend_license_db(self, file_name, oformat="JSON", default_no=False):
        return self.compatibility.extend_license_db(file_name, oformat, default_no)

    def simplify_license(self, expr):
        return self.license.simplify_license(expr)

    def get_license(self, expr):
        return self.license.get_license(expr)

    def choose_license(self, licenses):
        return self.license_chooser.choose(licenses)


# add to utils or similar easy to use function in future revision
def inbound_outbound_compatibility(outbound, inbounds, license_db=None):
    return LicenseCompatibilty(license_db).inbounds_outbound_compatibility(outbound, [inbounds])['compatibility']


if __name__ == '__main__':
    print(inbound_outbound_compatibility("MIT", "X11"))



###################################################################
#
# flict - FOSS License Compatibility Tool
#
# SPDX-FileCopyrightText: 2020 Henrik Sandklef
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
###################################################################

from flict.flictlib import logger
from flict.flictlib.format.format import FlictFormatter
from flict.flictlib.format.markdown_format import _compat_to_markdown
from flict.flictlib.return_codes import FlictError, ReturnCodes


class DotFormatter(FlictFormatter):

    def format_compats(self, compats):
        checked_set = set()
        result = []
        result.append("digraph depends {\n    node [shape=plaintext]\n")
        for compat in compats['compatibilities']:
            main_license = compat['license']
            for lic in compat['licenses']:
                inner_license = lic['license']
                if main_license == inner_license:
                    continue
                text_hash = _licenses_hash(main_license, inner_license)
                # If already handled, continue
                if text_hash in checked_set:
                    continue

                checked_set.add(text_hash)
                comp_left = lic['compatible_left']
                comp_right = lic['compatible_right']
                compat_dot = _compat_to_dot(
                    main_license, comp_left, inner_license, comp_right)
                if compat_dot != "":
                    result.append(f"    {compat_dot}\n")
        result.append('}\n')
        return " ".join(result)

# help functions


def _licenses_hash(a, b):
    separator = " "
    if a > b:
        return a + separator + b
    else:
        return b + separator + a


def _print_compare_line(left, right, color):
    return f'"{left}" -> "{right}"  {color}'


def _compat_to_dot(left, comp_left, right, comp_right):
    logger.main_logger.debug("_compat_to_dot")

    if comp_left == "true":
        logger.main_logger.debug("left true")
        if comp_right == "true":
            return _print_compare_line(left, right, '[dir=both] [color="darkgreen"]')
        if comp_right == "false":
            logger.main_logger.debug("1 dslkjsljdflskdjfljdf")
            res = _print_compare_line(left, right, '[color="black"]')
            logger.main_logger.debug(f'{left}    {right}')
            logger.main_logger.debug(f'dot:      {res}')
            logger.main_logger.debug(f'markdown: {_compat_to_markdown(None, comp_left, None, comp_right)}')
            return res

        if comp_right == "question" or comp_right == "undefined" or comp_right == "depends":
            return f"""
            {_print_compare_line(right, left, '[color="black"]')}
            {_print_compare_line(left, right, '[color="gray", style="dotted"]')}
            """
    elif comp_left == "false":
        logger.main_logger.debug("left false")

        if comp_right == "true":
            logger.main_logger.debug("left false right true")
            return _print_compare_line(right, left, '[color="black"]')
        if comp_right == "false":
            return f'"{left}"\n    "{right}"'
        if comp_right == "question" or comp_right == "undefined" or comp_right == "depends":
            return _print_compare_line(right, left, '[color="gray", style="dotted"]') + " \n "
    elif comp_left == "question" or comp_left == "undefined" or comp_left == "depends":
        logger.main_logger.debug("left QUD")
        # QUD---->
        if comp_right == "true":
            return f"""
            {_print_compare_line(left, right, '[color="black"]')}
            {_print_compare_line(right, left, '[color="gray", style="dotted"]')}
            """
        # QUD----|
        if comp_right == "false":
            return _print_compare_line(left, right, '[color="gray", style="dotted"]') + " \n "
        # QUD----Q|U|D
        if comp_right == "question" or comp_right == "undefined" or comp_right == "depends":
            return f"""
            {_print_compare_line(left, right, '[color="gray", style="dotted"]')}
            {_print_compare_line(right, left, '[color="gray", style="dotted"]')}
            """
    else:
        raise FlictError(ReturnCodes.RET_INTERNAL_ERROR,
                         f"Invalid state in dot_format: {left}, {comp_left}, {right}, {comp_right}")



###################################################################
#
# flict - FOSS License Compatibility Tool
#
# SPDX-FileCopyrightText: 2020 Henrik Sandklef
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
###################################################################


class FlictFormatter:

    def format_support_licenses(self, supported_licenses):
        return "default implementation | format_support_licenses(): " + str(supported_licenses)

    def format_license_list(self, license_list):
        return "default implementation | format_license_list(): " + str(license_list)

    def format_outbound_license(self, outbounds):
        return "default implementation | format_outbound_license(): " + str(outbounds)

    def format_license_combinations(self, combinations):
        return "default implementation | format_license_combinations(): " + str(combinations)

    def format_compats(self, compats):
        return "default implementation | format_compats(): " + str(compats)

    def format_simplified(self, simplified):
        return "default implementation | format_simplified(): " + str(simplified)

    def format_verified_license(self, license_expression, outbound_candidate):
        return "default implementation | format_verified_license(): " + str(license_expression)

    def format_compatibilities(self, compats):
        return

    def format_licenses(self, licenses):
        return

    def format_verification(self, verification):
        return

    def get_dep_license(self, dep, outbound):
        for lic_compat in dep.get('compatibility'):
            outbound_lic = lic_compat['outbound']['name']

            if outbound_lic == outbound:
                name = self.license.verified_to_license(lic_compat)
                return name
        return None

    def find_compat(self, compats, license_name):
        for compat in compats['compatibilities']:
            if compat["license"] == license_name:
                return compat

    def find_license_compat(self, compat, license_name):
        for lic in compat['licenses']:
            if lic['license'] == license_name:
                return lic['compatible_left'].replace("true", "Yes").replace("false", "No")



# SPDX-FileCopyrightText: 2022 Henrik Sandklef
#
# SPDX-License-Identifier: GPL-3.0-or-later

import json
import os
import re

from flict.flictlib.flict_config import DEFAULT_FLICT_ALIAS_FILE
from flict.flictlib.flict_config import VAR_DIR
from flict.flictlib.return_codes import FlictError, ReturnCodes

FLICT_ALIAS_FILE = os.path.join(VAR_DIR, DEFAULT_FLICT_ALIAS_FILE)


class Alias:

    def __init__(self, alias_file=None):
        try:
            with open(alias_file or FLICT_ALIAS_FILE) as fp:
                self.alias_object = json.load(fp)
        except BaseException:
            raise FlictError(ReturnCodes.RET_INVALID_ALIAS_FILE,
                             f'Could not open or parse alias file: {alias_file}')

    def _replace_aliases(self, lic):
        if not lic:
            return None
        for alias in self.alias_object['aliases']:
            if lic == alias['alias']:
                return alias['license']
        return lic

    def replace_aliases(self, license_expr):
        """Replaces license expression aliases (e.g GPLv2+) with supported
        license identifiers (e.g. GPL-2.0-or-later)

        Parameters:
            license_expr - a list of licenses

        """
        if not license_expr:
            return None

        # in case we have multiple spaces separating tokens, remove them
        license_expr = re.sub(r' +', ' ', license_expr)

        new_license_expr = []
        for lic in license_expr.replace(" WITH ", "_WITH_").split():
            new_license_expr.append(self._replace_aliases(lic.replace("_WITH_", " WITH ")))

        return " ".join(new_license_expr)

    def format_aliased(self, lic):
        alias = self.replace_aliases(lic)
        if alias != lic:
            return f'{lic} (aliased to {alias})'
        return lic



###################################################################
#
# flict - FOSS License Compatibility Tool
#
# SPDX-FileCopyrightText: 2021 Henrik Sandklef
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
###################################################################

from enum import Enum


class ReturnCodes(Enum):
    # common

    RET_SUCCESS = (0, "Success")
    RET_INTERNAL_ERROR = (1, "Internal error")
    # 2
    # 3
    # 4
    RET_MISSING_ARGS = (5, "Missing arguments")
    # 6
    # 7
    # 8
    # 9

    RET_INVALID_PROJECT = (10, "Invalid project")
    RET_INVALID_EXPRESSSION = (11, "Invalid expression")
    RET_FILE_NOT_FOUND = (12, "File not found")
    RET_INVALID_ALIAS_FILE = (13, "Invalid alias file")
    RET_INVALID_LICENSE_PREFERENCE = (13, "Invalid license preferences")
    RET_INVALID_MATRIX = (14, "Invalid matrix or matrix extension")

    @classmethod
    def get_help(cls, indent="  "):
        """
        Return help string for all defined enum values
        """
        ret = []
        for item in cls.__dict__.keys():
            if not item.startswith("RET_"):
                continue
            _v = getattr(cls, item)
            if not isinstance(_v, Enum):
                continue
            ret.append("{code}: {text}".format(code=_v.value[0], text=_v.value[1]))
        return indent + f"\n{indent}".join(ret)


class FlictError(Exception):

    def __init__(self, error_code, error_message=None):
        self._error_code = error_code
        if error_message is None:
            self._error_message = self._error_code.value[1]
        else:
            self._error_message = error_message

    def error_code(self):
        return self._error_code.value[0]

    def error_message(self):
        return self._error_message



###################################################################
#
# flict - FOSS License Compatibility Tool
#
# SPDX-FileCopyrightText: 2021 Henrik Sandklef, 2021 Konrad Weihmann
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
###################################################################

from flict.flictlib.format.format import FlictFormatter


class TextFormatter(FlictFormatter):

    def __init__(self):
        self.col_size = 18

    def format_support_licenses(self, supported_licenses):
        ret = [f" {item}" for item in supported_licenses]
        return "\n".join(ret)

    def format_license_list(self, license_list):
        return f"text implmentation | format_license_list(): {license_list}"

    def format_license_combinations(self, project):
        return str(project.projects_combinations())

    def format_outbound_license(self, outbound_candidates):
        return ", ".join(outbound_candidates)

    def format_simplified(self, simplified):
        return simplified['simplified']

    def format_verified_license(self, license_expression, outbound_candidate):
        ret_str = f'The licenses in the expression "{license_expression.strip()}" are'
        if len(outbound_candidate) == 0:
            ret_str += " not"
        ret_str += " compatible.\n"
        if len(outbound_candidate) == 0:
            ret_str += "No outbound license candidate could be identified due to license incompatibility."
        else:
            ret_str += "Outbound license candidates: "
            ret_str += ", ".join(outbound_candidate) + "\n"
            ret_str += "NOTE: the suggested outbound candidate licenses need to be manually reviewed."
        return ret_str

    def format_relicense_information(self, license_handler):
        ret = ["{lic} --> {later}".format(lic=rel.get('spdx', 'NOASSERTATION'), later=", ".join(rel.get('later', [])))
               for rel in license_handler.relicensing_information().get('original', {}).get('relicense_definitions', [])]
        return "\n".join(ret)

    def format_translation_information(self, license_handler):
        ret = []
        for transl_list in license_handler.translation_information():
            for transl in transl_list:
                ret.append(f"{transl} <--- {str(transl_list[transl])}")
        return "\n".join(ret)

    def format_policy_report(self, policy_report):
        outbounds = policy_report.get("policy_outbounds")
        policy_result = outbounds.get("policy_result")
        if policy_result == 0:
            status = "OK"
        elif policy_result == 1:
            status = "OK, with licenses to avoid"
        elif policy_result == 2:
            status = "Failed identifying outbound license"
        return f"""
        Status: {status}
        Allowed (suggested) outbound licenses: {outbounds.get("allowed", "")}
        Avoided (suggested) outbound licenses: {outbounds.get("avoid", "")}
        Denied (suggested) outbound licenses: {outbounds.get("denied", "")}
        """

    def _format_lic(self, lic):
        str_size = f'{0: <{self.col_size}}'
        return str_size.format(lic)[:self.col_size - 1] + " "

    def _format_line(self):
        return self._format_lic("-" * (self.col_size - 1))

    def _format_compats_licenses(self, compats, license_list):
        ret = []
        for lic in license_list:
            inner = []
            inner.append(self._format_lic(lic + ": "))
            compat = self.find_compat(compats, lic)
            for inner_lic in license_list:
                if inner_lic == lic:
                    lic_compat = "Yes"
                else:
                    lic_compat = self.find_license_compat(compat, inner_lic)
                inner.append(self._format_lic(lic_compat))
            ret.append("".join(inner))
        return ret

    def format_compats(self, compats):
        licenses = set()
        for compat in compats['compatibilities']:
            licenses.add(compat["license"])
        license_list = list(licenses)
        license_list.sort()

        ret = []

        inner = []
        inner.append(self._format_lic(""))
        for lic in license_list:
            inner.append(self._format_lic(lic))
        ret.append("".join(inner))

        ret.append("".join(self._format_line()))

        ret += self._format_compats_licenses(compats, license_list)

        return "\n".join(ret)

    def format_compatibilities(self, compats):
        compatible = compats['compatibility'] == "Yes"
        allowed = compats['allowed']
        return "Yes" if compatible and allowed else "No"



# SPDX-FileCopyrightText: 2022 Henrik Sandklef
#
# SPDX-License-Identifier: GPL-3.0-or-later

from flict.flictlib.alias import Alias
from flict.flictlib.license_parser import LicenseParserFactory
from flict.flictlib.return_codes import FlictError, ReturnCodes


class License():
    """Class managing license expressions, e.g.:
    X11
    MIT OR BSD-3-Clause
    GPL-2.0-or-later or (GPL-3.0-only WITH GCC-exception-3.1 AND curl
    """

    def __init__(self, denied_licenses, alias=None):
        self.alias = alias or Alias()
        self._denied_licenses = denied_licenses
        self.parser = LicenseParserFactory.get_parser()

    def get_license(self, expr):
        return self.parser.parse_license(expr)['license']

    def license_name(self, expr):
        return self.parser.license(expr)

    def operator(self, expr):
        return self.parser.operator(expr)

    def operands(self, expr):
        return self.parser.operands(expr)

    def licenses(self, expr):
        return self.parser.licenses(expr)

    def replace_aliases(self, expr):
        return self.alias.replace_aliases(expr)

    def denied_licenses(self):
        return self._denied_licenses

    def license_denied(self, license_):
        if self._denied_licenses:
            return license_ in self._denied_licenses
        return False

    def license_allowed(self, license_):
        return not self.license_denied(license_)

    def simplify_license(self, expr):
        try:
            aliased = self.alias.replace_aliases(expr)
            parsed = self.parser.parse_license([aliased])
            simplified = str(parsed['simplified'])
            return {
                "original": expr,
                "simplified": simplified,
            }
        except Exception as e:
            raise FlictError(ReturnCodes.RET_INVALID_EXPRESSSION, f"Could not parse or simplify license expression: {expr}. Cause: {str(e)}")

    def is_operator(self, expr):
        return self.parser.is_operator(expr)

    def is_license(self, expr):
        return self.parser.is_license(expr)

    def verified_to_license(self, parsed_expr):
        expr_type = parsed_expr['type']

        if expr_type == 'license':
            return parsed_expr['name']

        if expr_type == 'operator':
            name = parsed_expr['name']

            operands = []
            for op in parsed_expr['operands']:
                if op['compatibility'] == "Yes":
                    parsed_op = self.verified_to_license(op)
                    operands.append(parsed_op)
            operand_str = f" {name} "
            expr = operand_str.join(operands)
            if name == "OR" and len(operands) > 1:
                expr = f" ( {expr} ) "
            return expr



# SPDX-FileCopyrightText: 2021 Konrad Weihmann
#
# SPDX-License-Identifier: GPL-3.0-or-later

import setuptools
from flict.flictlib.flict_config import flict_version

with open("README.md") as i:
    _long_description = i.read()

requirements = []
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

requirements_dev = []
with open('requirements-dev.txt') as f:
    requirements_dev = f.read().splitlines()

setuptools.setup(
    name="flict",
    version=flict_version,
    author="Henrik Sanklef",
    author_email="hesa@sandklef.com",
    description="FOSS License Compatibility Tool",
    long_description=_long_description,
    long_description_content_type="text/markdown",
    license_files=('LICENSE',),
    url="https://github.com/vinland-technology/flict",
    packages=['flict', 'flict.flictlib', 'flict.flictlib.format', 'flict.flictlib.project'],
    entry_points={
        "console_scripts": [
            "flict = flict.__main__:main",
        ],
    },
    package_data={
        'flict': ['var/*'],
    },
    install_requires=requirements,
    extras_require={
        'dev': requirements_dev,
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Legal Industry",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Quality Assurance",
    ],
    python_requires='>=3.8',
)



# SPDX-FileCopyrightText: 2023 Henrik Sandklef
#
# SPDX-License-Identifier: GPL-3.0-or-later

import json
import logging

from enum import Enum
import osadl_matrix

from flict.flictlib.return_codes import FlictError, ReturnCodes
from flict.flictlib.alias import Alias


class CompatibilityStatus(Enum):
    LICENSE_COMPATIBILITY_COMPATIBLE = "Yes"
    LICENSE_COMPATIBILITY_INCOMPATIBLE = "No"
    LICENSE_COMPATIBILITY_UNKNOWN = "Unknown"
    LICENSE_COMPATIBILITY_MANUALLY_CHECK = "Check dependencies manually"
    LICENSE_COMPATIBILITY_UNDEFINED = "Undefined"


LICENSE_COMPATIBILITY_AND = "AND"
LICENSE_COMPATIBILITY_OR = "OR"
COMPATIBILITY_TAG = "compatibility"


class CompatibilityFactory:
    """Class to provide Compatibility objects via get_compatibility
    """

    @staticmethod
    def get_compatibility(alias=None, license_db=None):
        """Returns a Compatibility object.

            Parameters:
                alias: alias object
                licensedb: licensedb to use if not using default

        Currently only OsadlCompatibility is available.
        """
        return OsadlCompatibility(alias or Alias(), license_db)


class Compatibility:
    """Class to determine compatibility between licenses.

    This class need to be implemented in sub classes.
    """

    def __init__(self, alias, license_db=None):
        return None

    def check_compat(self, outbound, inbound):
        return None

    def supported_licenses(self):
        return None

    def display_compatibility(self):
        try:
            # build up license string from all expressions
            lic_str = " ".join(self._args.license_expression)

            # encode (flict) all the license expression
            lic_str = self._encode_license_expression(lic_str)

            # build up license string from the expression string
            _licenses = set()
            for lic in lic_str.split():

                lic_list = lic.replace("(", "").replace(
                    ")", "").replace(" ", "").replace("OR", " ").replace("AND", " ").strip().split(" ")

                for lic in lic_list:
                    if lic != "":
                        _licenses.add(self._decode_license_expression(lic))

            licenses = list(_licenses)

            inter_compats = self.check_compatibilities(licenses, self._args.extended_licenses)
        except BaseException:
            raise FlictError(ReturnCodes.RET_INVALID_EXPRESSSION,
                             f'Could not parse license expression: {self._args.license_expression}')

        return inter_compats

    def _encode_license_expression(self, license_expression):
        return license_expression.replace(" WITH ", "_WITH_")

    def _decode_license_expression(self, license_expression):
        return license_expression.replace("_WITH_", " WITH ")

    def check_compatibilities(self, licenses, check_all=False):
        """Check compatbilitiy between supplied licenses"""
        compats = []

        if check_all:
            supported = self.supported_licenses()
            if 'Compatibility' in supported:
                supported.remove('Compatibility')

            licenses_set = set(licenses + supported)
            outer_licenses = list(licenses_set)
        else:
            outer_licenses = set(licenses)

        for lic_a in outer_licenses:
            inner_licenses = []
            for lic_b in licenses:

                comp_left = self.check_compat(lic_a, lic_b)['compatibility']
                comp_right = self.check_compat(lic_b, lic_a)['compatibility']

                if CompatibilityStatus.LICENSE_COMPATIBILITY_UNKNOWN.value in (comp_left, comp_right):
                    supported = self.supported_licenses()
                    lic_bad = ",".join({lic for lic in (lic_a, lic_b) if lic not in supported})
                    raise FlictError(ReturnCodes.RET_INVALID_EXPRESSSION,
                                     f'License expression "{lic_bad}" is not supported.')

                inner_licenses.append({
                    'license': lic_b,
                    'compatible_right': self._compatibility_status_json(comp_right),
                    'compatible_left': self._compatibility_status_json(comp_left),
                })

            compats.append({
                'license': lic_a,
                'licenses': inner_licenses,
            })

        return {
            'compatibilities': compats,
        }

    def _compatibility_status_json(self, status):
        if status == "Yes":
            return "true"
        elif status == "No":
            return "false"
        elif status == "Undefined":
            return "undefined"
        elif status == "Check dependencies manually":
            return "depends"
        elif status == "Unknown":
            return "question"

    def extend_license_db(self, file_name):
        return None


class OsadlCompatibility(Compatibility):
    """Class to determine compatibility between licenses using OSADL's matrix

    This class implements Compatibility
    """

    def __init__(self, alias, license_db=None):
        self.license_db = license_db
        self.alias = alias

    def check_compat(self, _outbound, _inbound):

        outbound = self.alias.replace_aliases(_outbound)
        inbound = self.alias.replace_aliases(_inbound)

        supported = osadl_matrix.supported_licenses()
        if outbound not in supported:
            raise FlictError(ReturnCodes.RET_INVALID_EXPRESSSION,
                             f'Compatibility between \"{outbound}\" and \"{inbound}\" could not be determined, since \"{outbound}\" is an unknown license')
        if inbound not in supported:
            raise FlictError(ReturnCodes.RET_INVALID_EXPRESSSION,
                             f'Compatibility between \"{outbound}\" and \"{inbound}\" could not be determined, since \"{inbound}\" is an unknown license')

        raw_result = osadl_matrix.get_compatibility(outbound, inbound, self.license_db)

        logging.info(f"check_compat({outbound}, {inbound} => {raw_result}")

        if raw_result == osadl_matrix.OSADLCompatibility.YES:
            result = CompatibilityStatus.LICENSE_COMPATIBILITY_COMPATIBLE.value
        elif raw_result == osadl_matrix.OSADLCompatibility.NO:
            result = CompatibilityStatus.LICENSE_COMPATIBILITY_INCOMPATIBLE.value
        elif raw_result == osadl_matrix.OSADLCompatibility.UNKNOWN:
            result = CompatibilityStatus.LICENSE_COMPATIBILITY_UNKNOWN.value
            logging.debug(f"compatibility: {outbound}  --> {inbound}: {result}")
        elif raw_result == osadl_matrix.OSADLCompatibility.CHECKDEP:
            result = CompatibilityStatus.LICENSE_COMPATIBILITY_MANUALLY_CHECK.value
            logging.debug(f"compatibility: {outbound}  --> {inbound}: {result}")
        elif raw_result == osadl_matrix.OSADLCompatibility.UNDEF:
            raise FlictError(ReturnCodes.RET_INVALID_EXPRESSSION,
                             f"Compatibility between \"{outbound}\" and \"{inbound}\" could not be determined. The result was: {raw_result}")

        return {
            "inbound": inbound,
            "outbound": outbound,
            COMPATIBILITY_TAG: result,
        }

    def supported_licenses(self):
        """Returns a list of supported licenses"""
        return list(osadl_matrix.supported_licenses(self.license_db))

    def _create_matrix(self, raw):
        """
        Copied from https://github.com/priv-kweihmann/osadl-matrix/blob/master/scripts/scrapper.py
        Modified by Henrik Sandklef
        # SPDX-FileCopyrightText: 2021 Konrad Weihmann
        # SPDX-FileCopyrightText: 2022 Henrik Sandklef
        # SPDX-License-Identifier: Unlicensed
        """
        # create a csv

        _csv = [
            ['Compatibility*'] + sorted(raw.keys()),
        ]

        for key in sorted(raw.keys()):
            _line = [key]
            if key.startswith('timeformat') or key.startswith('timestamp'):
                continue
            for key2 in sorted(raw.keys()):
                try:
                    _line.append(raw[key][key2])
                except KeyError:
                    _line.append("Unknown")
            _csv.append(_line)

        for line in _csv:
            if len(line) < 3:
                continue
        return _csv

    def _create_matrix_csv_data(self, file_name, default_no=True):
        # read file with additional license data
        with open(file_name) as fp:
            additional_data = json.load(fp)['osadl_additional_licenses']

        # read osadl matrix data
        with open(osadl_matrix.OSADL_MATRIX_JSON) as fp:
            osadl_data = json.load(fp)

        # merge osadl and our data
        for key, value in additional_data.items():
            if key in osadl_data:
                osadl_data[key].update(value)
            else:
                osadl_data[key] = value

        new_matrix_data = self._create_matrix(osadl_data)
        rows = []
        for row in new_matrix_data:
            rows.append(",".join(row))
        return "\n".join(rows)

    def __add_no_to_missing_variables(self, osadl_data):
        all_keys = osadl_data.keys()
        fixed_matrix = osadl_data.copy()
        for outer_key in all_keys:
            fixed_matrix[outer_key] = osadl_data[outer_key].copy()
            for inner_key in osadl_data[outer_key].keys():
                for every_key in all_keys:
                    if inner_key not in fixed_matrix:
                        raise FlictError(ReturnCodes.RET_INVALID_MATRIX,
                                         f'Can\'t add "{every_key}" to "{inner_key}" since "{inner_key}" does not exist in matrix')
                    if every_key not in fixed_matrix[inner_key]:
                        fixed_matrix[inner_key][every_key] = "No"
        return fixed_matrix

    def _create_matrix_json_data(self, file_name, default_no=False):

        # read file with additional license data
        with open(file_name) as fp:
            additional_data = json.load(fp)['osadl_additional_licenses']

        # read osadl matrix data
        with open(self.license_db) as fp:
            osadl_data = json.load(fp)

        # merge each additional entry's value with the existing
        for key, value in additional_data.items():
            if key not in osadl_data:
                osadl_data[key] = {}
            osadl_data[key].update(value)

        timestamp = osadl_data.get("timestamp", "")
        timeformat = osadl_data.get("timeformat", "")
        osadl_data.pop("timeformat", None)
        osadl_data.pop("timestamp", None)

        # Check if we should add "No" where no values if is provided in additional matrix
        if default_no:
            osadl_data = self.__add_no_to_missing_variables(osadl_data)

        # Check matrix completeness
        violations = [k for k, v in osadl_data.items() if not all(b in v for b in osadl_data.keys())]
        valid = not any(violations)
        if not valid:
            raise FlictError(ReturnCodes.RET_INVALID_MATRIX,
                             f'Set of keys differs from set of values, see "{violations}". Make sure your additional matrix is complete')

        osadl_data['timeformat'] = timeformat
        osadl_data['timestamp'] = timestamp

        return json.dumps(osadl_data, indent=4)

    def extend_license_db(self, file_name, oformat="JSON", default_no=True):
        """Extends the current license db with licese db provided in file_name.

        Parameters:
            file_name - file (str) with license db (JSON)

        See SETTINGS.md for more information about format and
        requirements for a custom database.
        """
        if oformat.lower() == "csv":
            return self._create_matrix_csv_data(file_name)

        return self._create_matrix_json_data(file_name, default_no)


class LicenseChooser:
    """Interface for classes to choose from inbound licenses (where a
    choice is offered via OR).

    """

    def __init__(self, licenses):
        """Parameters:
               licenses - the licenses to create a choose object for
        """
        self.licenses = licenses

    def choose(self, licenses):
        """Choose the most preferred license

        Parameters:
            licenses - list of licenses to choose the most preferred from
        """
        index = None
        # Loop through the supplied list of licenses, typically the
        # set of all licenses in a project
        for lic in licenses:
            # If the license is not found in the license preference list,
            # raise an error
            if lic not in self.licenses:
                raise FlictError(ReturnCodes.RET_INVALID_LICENSE_PREFERENCE,
                                 f'The supplied license preference list is incomplete. The license \"{lic}\" is missing from {self.licenses}.')

            # identify the index and remember the lowest index (most preferred)
            lic_index = self.licenses.index(lic)
            if index is None or lic_index < index:
                index = lic_index

        if index is None:
            return None
        else:
            return self.licenses[index]

    def list_licenses(self):
        """returns a list of licenses in preference order"""
        return self.licenses


class CustomLicenseChooser(LicenseChooser):
    """This class provides a custom way to choose from inbound licenses
    (where a choice is offered via OR). By providing a list of licenses in order of preference, this list is used"""
    pass


class CompatibilityLicenseChooser(LicenseChooser):
    """This class provides a simple way to choose from inbound licenses
    (where a choice is offered via OR).

    It counts how many other licenses each license is compatible
    with. The more licenses a license is compatible with the more
    preferred it will be. If two licenses have the same number of
    compaitbilities alpabetical order will be used to choose license.
    """

    def __init__(self, licenses):
        """Parameters:
               licenses - the licenses to create a chooser object for
        """
        self.licenses = self._license_preferences(licenses)

    def _count_compat(self, inbound, supported_licenses):
        compat_cnt = 0
        for outbound in supported_licenses:
            if osadl_matrix.is_compatible(outbound, inbound):
                compat_cnt += 1
        return {
            "inbound": inbound,
            "compatible_count": compat_cnt,
        }

    def _count_compats(self, supported_licenses):
        compats = []
        for inbound in supported_licenses:
            compats.append(self._count_compat(inbound, supported_licenses))
        compats_sorted = sorted(compats, key=lambda element: (element['compatible_count'], element['inbound']))
        compats_sorted.reverse()
        return compats_sorted

    def _license_preferences(self, supported_licenses):
        compats_sorted = self._count_compats(supported_licenses)
        pref_list = []
        for lic in compats_sorted:
            pref_list.append(lic['inbound'])

        return pref_list



###################################################################
#
# flict - FOSS License Compatibility Tool
#
# SPDX-FileCopyrightText: 2020 Henrik Sandklef
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
###################################################################

from flict.flictlib.format.json_format import JsonFormatter
from flict.flictlib.format.text_format import TextFormatter
from flict.flictlib.format.dot_format import DotFormatter
from flict.flictlib.format.markdown_format import PackageMarkdownFlictFormatter


class FormatterFactory:

    _instance = None

    @staticmethod
    def formatter(format_):
        if FormatterFactory._instance is None:
            if format_.lower() == "json":
                FormatterFactory._instance = JsonFormatter()
            elif format_.lower() == "text":
                FormatterFactory._instance = TextFormatter()
            elif format_.lower() == "dot":
                FormatterFactory._instance = DotFormatter()
            elif format_.lower() == "markdown":
                FormatterFactory._instance = PackageMarkdownFlictFormatter()
            # else ignored
        # else ignored
        return FormatterFactory._instance
